﻿/*
  TAN Placeholder Plugin
  Copyright (C) 2015 Nikolaus Hammler <niki@hammler.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Drawing;
using System.Text.RegularExpressions;

using KeePass.Plugins;
using KeePass.Util;
using KeePass.Util.Spr;

using KeePassLib;
using KeePassLib.Collections;
using KeePassLib.Security;
using KeePassLib.Utility;

namespace TAN_Placeholder
{
    public sealed class TAN_PlaceholderExt : Plugin
    {
        private const string UPDATEURL = "https://secure.nobaq.net/keepass-updates.txt";
        private const string m_placeholderUuid = "{USETAN:I:group-uuid:optional tan nr}";
        private const string m_placeholderTitle = "{USETAN:T:group-name:optional tan nr}";

        private IPluginHost m_host = null;

        public override bool Initialize(IPluginHost host)
        {
            if (host == null)
                return false;

            this.m_host = host;

            SprEngine.FilterPlaceholderHints.Add(m_placeholderUuid);
            SprEngine.FilterPlaceholderHints.Add(m_placeholderTitle);
            AutoType.FilterCompilePre += this.OnAutoTypeFilterCompilePre;

            return true;
        }

        public override void Terminate()
        {
            SprEngine.FilterPlaceholderHints.Remove(m_placeholderUuid);
            SprEngine.FilterPlaceholderHints.Remove(m_placeholderTitle);
            AutoType.FilterCompilePre -= this.OnAutoTypeFilterCompilePre;
        }

        public override Image SmallIcon 
        {
            get { return (Image)KeePass.Program.Resources.GetObject("B16x16_Key_New"); }
        }

        public override string UpdateUrl
        {
            get { return UPDATEURL; }
        }

        private void OnAutoTypeFilterCompilePre(object sender, AutoTypeEventArgs autoTypeEventArgs)
        {
            String pattern;
            pattern = Regex.Escape("{") + "USETAN:(I|T):([^:}]+)(:([0-9]+))?" + Regex.Escape("}");

            Regex replacer = new Regex(pattern, RegexOptions.IgnoreCase);

            autoTypeEventArgs.Sequence = replacer.Replace(autoTypeEventArgs.Sequence, match =>
            {
                if (match.Groups.Count != 5)
                {
                    // error in regex
                    // Do not replace anything by returning the original value
                    // Let KeePass display the appropriate message
                    return match.Value;
                }

                String matchedType = match.Groups[1].ToString().ToUpper();
                String matchedKey = match.Groups[2].ToString();
                String matchedTanId = match.Groups[4].ToString();

                PwGroup tanGrp = null;
                if (matchedType.CompareTo("I") == 0)
                {
                    PwUuid tanGrpUuid = null;
                    try
                    {
                        tanGrpUuid = new PwUuid(MemUtil.HexStringToByteArray(matchedKey));
                    }
                    catch (Exception)
                    {
                        // cannot convert I parameter (UUID), is not hex
                        // invalid format, return original value
                        return match.Value;
                    }
                    tanGrp = m_host.Database.RootGroup.FindGroup(tanGrpUuid, true);
                }
                else if (matchedType.CompareTo("T") == 0)
                {
                    PwObjectList<PwGroup> groups = autoTypeEventArgs.Database.RootGroup.GetGroups(true);
                    foreach (PwGroup grp in groups)
                    {
                        if (grp.Name.CompareTo(matchedKey) == 0)
                        {
                            tanGrp = m_host.Database.RootGroup.FindGroup(grp.Uuid, true);
                            break;
                        }
                    }
                }
                else
                {
                    // First argument is neither I or T
                    // return original value
                    return match.Value;
                }

                if (tanGrp == null)
                {
                    m_host.MainWindow.EnsureVisibleForegroundWindow(true, true);
                    throw new Exception("{USETAN}: Supplied group was not found");
                }

                foreach (PwEntry entry in tanGrp.Entries)
                {
                    if (PwDefs.IsTanEntry(entry) && !entry.Expires)
                    {
                        bool ourMatch = true;
                        if (matchedTanId.Length > 0)
                        {
                            if (entry.Strings.Get(PwDefs.TanIndexField).ReadString().CompareTo(matchedTanId) != 0)
                            {
                                ourMatch = false;
                            }
                        }

                        if (ourMatch)
                        {
                            EntryUtil.ExpireTanEntryIfOption(entry, autoTypeEventArgs.Database);
                            return entry.Strings.Get(PwDefs.PasswordField).ReadString();
                        }
                    }
                }

                m_host.MainWindow.EnsureVisibleForegroundWindow(true, true);
                if (matchedTanId.Length > 0)
                    throw new Exception("{USETAN}: Requested TAN entry was not found or has been used already");
                else
                    throw new Exception("{USETAN}: No unused TAN available");
            });
        }
    }
}
